open User

table rootAdmin : { User : user }
  PRIMARY KEY User,
  CONSTRAINT User FOREIGN KEY User REFERENCES user(Id)

task initialize = fn () =>
    b <- nonempty rootAdmin;
    if b then
        return ()
    else
        dml (INSERT INTO rootAdmin (User) VALUES ({[root]}))

type forum = int
sequence forumIds
table forum : { Id : int, Nam : string, Public : bool }
  PRIMARY KEY Id,
  CONSTRAINT Nam UNIQUE Nam

policy sendClient (SELECT forum.Id, forum.Public
                   FROM forum)

policy sendClient (SELECT *
                   FROM forum, rootAdmin, user
                   WHERE rootAdmin.User = user.Id
                     AND known(user.Pass))

policy sendClient (SELECT *
                   FROM forum
                   WHERE forum.Public)

policy mayInsert (SELECT * 
                  FROM forum AS New, rootAdmin, user
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass)
                    AND known(New.Nam))

policy mayUpdate (SELECT * 
                  FROM forum AS New, forum AS Old, rootAdmin, user
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass)
                    AND known(New.Nam)
                    AND New.Id = Old.Id)

policy mayDelete (SELECT * 
                  FROM forum AS Old, rootAdmin, user
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass))

type level = int
val read = 1
val write = 2
val admin = 3

table acl : { Forum : forum, User : user, Level : level }
  PRIMARY KEY (Forum, User),
  CONSTRAINT Forum FOREIGN KEY Forum REFERENCES forum(Id),
  CONSTRAINT User FOREIGN KEY User REFERENCES user(Id),
  CONSTRAINT Level CHECK 1 <= Level AND Level <= 3

policy mayInsert (SELECT *
                  FROM acl AS New, user, rootAdmin
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass))

policy mayInsert (SELECT *
                  FROM acl AS New, user, acl AS You
                  WHERE New.Forum = You.Forum
                    AND You.User = user.Id
                    AND known(user.Pass)
                    AND You.Level >= {[admin]})

policy mayUpdate (SELECT *
                  FROM acl AS New, acl AS Old, user, rootAdmin
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass))

policy mayUpdate (SELECT *
                  FROM acl AS New, acl AS Old, user, acl AS You
                  WHERE New.Forum = You.Forum
                    AND Old.Forum = You.Forum
                    AND You.User = user.Id
                    AND known(user.Pass)
                    AND You.Level >= {[admin]})

policy mayDelete (SELECT *
                  FROM acl AS Old, user, rootAdmin
                  WHERE rootAdmin.User = user.Id
                    AND known(user.Pass))

policy mayDelete (SELECT *
                  FROM acl AS Old, user, acl AS You
                  WHERE Old.Forum = You.Forum
                    AND You.User = user.Id
                    AND known(user.Pass)
                    AND You.Level >= {[admin]})

policy sendClient (SELECT *
                   FROM acl, user
                   WHERE acl.User = user.Id
                     AND known(user.Pass))

policy sendClient (SELECT *
                   FROM acl, user, rootAdmin
                   WHERE user.Id = rootAdmin.User
                     AND known(user.Pass))

policy sendClient (SELECT *
                   FROM acl AS Other, user, acl AS You
                   WHERE You.User = user.Id
                     AND known(user.Pass)
                     AND Other.Forum = You.Forum
                     AND You.Level >= {[admin]})

policy sendClient (SELECT *
                   FROM forum, acl, user
                   WHERE acl.Forum = forum.Id
                     AND acl.User = user.Id
                     AND known(user.Pass))

type message = int
sequence messageIds
table message : { Id : message, Forum : forum, User : user, Subject : string, Body : string, Time : time }
  PRIMARY KEY Id,
  CONSTRAINT Forum FOREIGN KEY Forum REFERENCES forum(Id),
  CONSTRAINT User FOREIGN KEY User REFERENCES user(Id)

policy sendClient (SELECT message.Id, message.Forum
                   FROM message)

policy sendClient (SELECT *
                   FROM message, forum
                   WHERE message.Forum = forum.Id
                     AND forum.Public)

policy sendClient (SELECT *
                   FROM message, rootAdmin, user
                   WHERE rootAdmin.User = user.Id
                     AND known(user.Pass))

policy sendClient (SELECT *
                   FROM message, acl, user
                   WHERE acl.Forum = message.Forum
                     AND acl.User = user.Id
                     AND known(user.Pass))

policy mayInsert (SELECT *
                  FROM message AS New, user, rootAdmin
                  WHERE New.User = user.Id
                    AND user.Id = rootAdmin.User
                    AND known(user.Pass)
                    AND known(New.Subject)
                    AND known(New.Body))

policy mayInsert (SELECT *
                  FROM message AS New, user, acl
                  WHERE New.User = user.Id
                    AND New.Forum = acl.Forum
                    AND user.Id = acl.User
                    AND known(user.Pass)
                    AND acl.Level >= {[write]}
                    AND known(New.Subject)
                    AND known(New.Body))

policy mayDelete (SELECT *
                  FROM message AS Old, user, rootAdmin
                  WHERE user.Id = rootAdmin.User
                    AND known(user.Pass))

policy mayDelete (SELECT *
                  FROM message AS Old, user, acl
                  WHERE Old.Forum = acl.Forum
                    AND user.Id = acl.User
                    AND known(user.Pass)
                    AND acl.Level >= {[admin]})

policy mayUpdate (SELECT *
                  FROM message AS New, message AS Old, user, rootAdmin
                  WHERE New.Id = Old.Id
                    AND Old.User = New.User
                    AND user.Id = rootAdmin.User
                    AND known(user.Pass)
                    AND known(New.Subject)
                    AND known(New.Body))

policy mayUpdate (SELECT *
                  FROM message AS New, message AS Old, user, acl
                  WHERE New.Id = Old.Id
                    AND Old.User = New.User
                    AND New.Forum = acl.Forum
                    AND Old.Forum = New.Forum
                    AND user.Id = acl.User
                    AND known(user.Pass)
                    AND acl.Level >= {[write]}
                    AND known(New.Subject)
                    AND known(New.Body))

val forceUserId =
    u <- User.userId ();
    case u of
        None => error <xml>You must log in to proceed further.</xml>
      | Some u => return u

val haveRoot =
    u <- forceUserId;
    oneRowE1 (SELECT COUNT( * ) > 0
              FROM rootAdmin
              WHERE rootAdmin.User = {[u]})

val checkRoot =
    b <- haveRoot;
    if b then
        return ()
    else
        error <xml>You need root admin privileges to do that.</xml>

fun mayRead id =
    b <- oneRowE1 (SELECT (forum.Public)
                   FROM forum
                   WHERE forum.Id = {[id]});
    if b then
        return True
    else
        u <- User.userId ();
        case u of
            None => return False
          | Some u =>
            b <- haveRoot;
            if b then
                return True
            else
                oneRowE1 (SELECT COUNT( * ) > 0
                          FROM acl
                          WHERE acl.Forum = {[id]}
                            AND acl.User = {[u]})

fun checkRead id =
    b <- mayRead id;
    if b then
        return ()
    else
        error <xml>You don't have read access to this forum.</xml>

fun mayWrite id =
    u <- User.userId ();
    case u of
        None => return False
      | Some u =>
        b <- haveRoot;
        if b then
            return True
        else
            oneRowE1 (SELECT COUNT( * ) > 0
                      FROM acl
                      WHERE acl.Forum = {[id]}
                        AND acl.User = {[u]}
                        AND acl.Level >= {[write]})

fun checkWrite id =
    b <- mayWrite id;
    if b then
        return ()
    else
        error <xml>You don't have write access to this forum.</xml>

fun mayAdmin id =
    u <- User.userId ();
    case u of
        None => return False
      | Some u =>
        b <- haveRoot;
        if b then
            return True
        else
            oneRowE1 (SELECT COUNT( * ) > 0
                      FROM acl
                      WHERE acl.Forum = {[id]}
                        AND acl.User = {[u]}
                        AND acl.Level >= {[admin]})

fun checkAdmin id =
    b <- mayAdmin id;
    if b then
        return ()
    else
        error <xml>You don't have admin access to this forum.</xml>

fun wrap title body =
    bl <- User.blurb ();
    return <xml>
      <head>
        <title>Forums - {[title]}</title>
      </head>

      <body>
        {bl}

        <h1>{[title]}</h1>

        {body}
      </body>
    </xml>

fun read id =
    let
        fun getFid () = oneRowE1 (SELECT (message.Forum)
                                  FROM message
                                  WHERE message.Id = {[id]})

        fun modifyMessage r =
            fid <- getFid ();
            checkAdmin fid;

            dml (UPDATE message
                 SET Subject = {[r.Subject]}, Body = {[r.Body]}
                 WHERE Id = {[id]});

            read id
    in
        fid <- getFid ();
        checkRead fid;
        mw <- mayWrite fid;
        ma <- mayAdmin fid;

        msg <- oneRow (SELECT forum.Nam, message.Subject, message.Body, message.Time, user.Nam
                       FROM message, forum, user
                       WHERE message.Id = {[id]}
                         AND forum.Id = message.Forum
                         AND user.Id = message.User);

        wrap (msg.Forum.Nam ^ " - " ^ msg.Message.Subject) <xml>
          Posted by: {[msg.User.Nam]}<br/>
          Posted at: {[msg.Message.Time]}<br/>
          <br/>
          {[msg.Message.Body]}
          
          {if mw then
               <xml>
                 <h2>Post a reply</h2>

                 <form><table>
                   <tr> <th>Subject:</th> <td><textbox{#Subject} value={if String.isPrefix {Prefix = "Re:",
                                                                                            Full = msg.Message.Subject} then
                                                                            msg.Message.Subject
                                                                        else
                                                                            "Re: " ^ msg.Message.Subject}/></td> </tr>
                   <tr> <th>Body</th> <td><textarea{#Body}/></td> </tr>
                   <tr> <td><submit action={post fid}/></td> </tr>
                 </table></form>
               </xml>
           else
               <xml/>}

          {if ma then
               <xml>
                 <h2>Admin actions</h2>

                 <a link={deleteMessage id}>[Delete message]</a>

                 <h3>Modify message</h3>

                 <form><table>
                   <tr> <th>Subject:</th> <td><textbox{#Subject} value={msg.Message.Subject}/></td> </tr>
                   <tr> <th>Body</th> <td><textarea{#Body}>{[msg.Message.Body]}</textarea></td> </tr>
                   <tr> <td><submit action={modifyMessage}/></td> </tr>
                 </table></form>
               </xml>
           else
               <xml/>}
        </xml>
    end

and post fid r =
    checkWrite fid;
    u <- forceUserId;
    
    id <- nextval messageIds;
    dml (INSERT INTO message (Id, Forum, Subject, Body, User, Time)
         VALUES ({[id]}, {[fid]}, {[r.Subject]}, {[r.Body]}, {[u]}, CURRENT_TIMESTAMP));

    messages fid

and messages id =
    let
        fun doPerm uid level =
            checkAdmin id;

            (case level of
                 "0" => dml (DELETE FROM acl
                             WHERE Forum = {[id]}
                               AND User = {[uid]})
               | s =>
                 b <- oneRowE1 (SELECT COUNT( * ) > 0
                                FROM acl
                                WHERE acl.Forum = {[id]}
                                  AND acl.User = {[uid]});
                 if b then
                     dml (UPDATE acl
                          SET Level = {[readError s]}
                          WHERE Forum = {[id]}
                            AND User = {[uid]})
                 else
                     dml (INSERT INTO acl (Forum, User, Level)
                          VALUES ({[id]}, {[uid]}, {[readError s]})));
            
            messages id

        fun permById uid r = doPerm uid r.Level

        fun permByName r =
            uid <- oneOrNoRowsE1 (SELECT (user.Id)
                                  FROM user
                                  WHERE user.Nam = {[r.Nam]});
            case uid of
                None => error <xml>Username not found</xml>
              | Some uid => doPerm uid r.Level
    in
        checkRead id;
        mw <- mayWrite id;
        ma <- mayAdmin id;

        name <- oneRowE1 (SELECT (forum.Nam)
                          FROM forum
                          WHERE forum.Id = {[id]});

        msgs <- queryX (SELECT user.Nam, message.Id, message.Subject, message.Time
                        FROM message, user
                        WHERE message.User = user.Id
                          AND message.Forum = {[id]}
                        ORDER BY message.Time DESC)
                       (fn r => <xml><tr>
                         <td><a link={read r.Message.Id}>{[r.Message.Subject]}</a></td>
                         <td>{[r.User.Nam]}</td>
                         <td>{[r.Message.Time]}</td>
                         {if ma then
                              <xml><td><a link={deleteMessage r.Message.Id}>[Delete]</a></td></xml>
                          else
                              <xml/>}
                       </tr></xml>);

        perms <- (if ma then
                      queryX (SELECT user.Id, user.Nam, acl.Level
                              FROM user, acl
                              WHERE acl.Forum = {[id]}
                                AND acl.User = user.Id)
                             (fn r => <xml><tr><form>
                               <td>{[r.User.Nam]}</td>
                               <td><select{#Level}>
                                 <option value="0">None</option>
                                 <option value="1" selected={r.Acl.Level = 1}>Read</option>
                                 <option value="2" selected={r.Acl.Level = 2}>Write</option>
                                 <option value="3" selected={r.Acl.Level = 3}>Admin</option>
                               </select></td>
                               <td><submit value="Set" action={permById r.User.Id}/></td>
                             </form></tr></xml>)
                  else
                      return <xml/>);

        wrap name <xml>
          <table>
            <tr> <th>Subject</th> <th>User</th> <th>Time</th> </tr>
            {msgs}
          </table>

          {if mw then
               <xml>
                 <h2>Post a new message</h2>

                 <form><table>
                   <tr> <th>Subject:</th> <td><textbox{#Subject}/></td> </tr>
                   <tr> <th>Body</th> <td><textarea{#Body}/></td> </tr>
                   <tr> <td><submit action={post id}/></td> </tr>
                 </table></form>
               </xml>
           else
               <xml/>}

          {if ma then
               <xml>
                 <h2>Manage permissions</h2>

                 <table>
                   <tr> <th>User</th> <th>Level</th> </tr>
                   {perms}
                   <tr><form> <td><textbox{#Nam}/></td> <td><select{#Level}>
                     <option value="1">Read</option>
                     <option value="2">Write</option>
                     <option value="3">Admin</option>
                   </select></td>
                   <td><submit value="Set" action={permByName}/></td>
                   </form></tr>
                 </table>
               </xml>
           else
               <xml/>}
        </xml>
    end

and deleteMessage id =
    let
        fun check () = 
            fid <- oneRowE1 (SELECT (message.Forum)
                             FROM message
                             WHERE message.Id = {[id]});
            checkAdmin fid;
            return fid

        fun reallyDeleteMessage () =
            fid <- check ();
            dml (DELETE FROM message
                        WHERE Id = {[id]});
            messages fid
    in
        fid <- check ();
        
        r <- oneRow (SELECT user.Nam, message.Subject
                     FROM message, user
                     WHERE message.Id = {[id]}
                       AND message.User = user.Id);

        wrap "Confirm message deletion" <xml>
          <p>Are you sure you want to delete the message posted by <tt>{[r.User.Nam]}</tt> with subject <tt>{[r.Message.Subject]}</tt>?</p>

          <form> <submit value="Yes, delete it!" action={reallyDeleteMessage}/> </form>
        </xml>
    end

fun main () =
    u <- User.userId ();
    case u of
        None =>
        forums <- queryX (SELECT forum.Id AS Id, forum.Nam AS Nam
                          FROM forum
                          WHERE forum.Public)
                         (oneForum False);
        
        wrap "Index" <xml>
          <table>
            <tr> <th>Forum</th> </tr>
            {forums}
          </table>
        </xml>
      | Some u =>
        root <- haveRoot;
        
        forums <- queryX (SELECT forum.Id AS Id, forum.Nam AS Nam
                          FROM forum
                          WHERE forum.Public
                          UNION SELECT forum.Id AS Id, forum.Nam AS Nam
                                FROM forum, rootAdmin
                                WHERE rootAdmin.User = {[u]}
                          UNION SELECT forum.Id AS Id, forum.Nam AS Nam
                                FROM forum, acl
                                WHERE acl.Forum = forum.Id
                                  AND acl.User = {[u]}
                          ORDER BY Nam)
                         (oneForum root);
    
        wrap "Index" <xml>
          <table>
            <tr> <th>Forum</th> </tr>
            {forums}
          </table>

          {if root then
               <xml>
                 <h2>Add a new forum</h2>
                 
                 <form><table>
                   <tr> <th>Name:</th> <td><textbox{#Nam}/></td> </tr>
                   <tr> <th>Public?</th> <td><checkbox{#Public}/></td> </tr>
                   <tr> <td><submit action={newForum}/></td> </tr>
                 </table></form>
               </xml>
           else
               <xml/>}
        </xml>

and oneForum amRoot r = <xml><tr>
  <td><a link={messages r.Id}>{[r.Nam]}</a></td>
  {if amRoot then
       <xml>
         <td><a link={editForum r.Id}>[Edit]</a></td>
         <td><a link={deleteForum r.Id}>[Delete]</a></td>
       </xml>
   else
       <xml/>}
</tr></xml>

and newForum r =
    checkRoot;
    id <- nextval forumIds;
    dml (INSERT INTO forum (Id, Nam, Public)
         VALUES ({[id]}, {[r.Nam]}, {[r.Public]}));
    main ()

and editForum id =
    let
        fun saveForum r =
            checkRoot;
            dml (UPDATE forum SET Nam = {[r.Nam]}, Public = {[r.Public]}
                 WHERE Id = {[id]});
            main ()
    in
        checkRoot;

        r <- oneRow1 (SELECT *
                      FROM forum
                      WHERE forum.Id = {[id]});

        wrap ("Editing forum \"" ^ r.Nam ^ "\"") <xml>
          <form><table>
            <tr> <th>New name:</th> <td><textbox{#Nam} value={r.Nam}/></td> </tr>
            <tr> <th>Public?</th> <td><checkbox{#Public} checked={r.Public}/></td> </tr>
            <tr> <td><submit action={saveForum}/></td> </tr>
          </table></form>
        </xml>
    end

and deleteForum id =
    let
        fun reallyDeleteForum () =
            checkRoot;
            dml (DELETE FROM forum WHERE Id = {[id]});
            main ()
    in
        checkRoot;

        oldName <- oneRowE1 (SELECT (forum.Nam)
                             FROM forum
                             WHERE forum.Id = {[id]});

        wrap ("Deleting forum \"" ^ oldName ^ "\"") <xml>
          <form>
            <submit value="I'm sure I want to delete this." action={reallyDeleteForum}/>
          </form>
        </xml>
    end
