type poll = int
sequence pollIds
table poll : { Id : poll, Nam : string, Code : int, Live : bool }
  PRIMARY KEY Id

policy mayInsert (SELECT *
                  FROM poll AS New
                  WHERE known(New.Nam)
                    AND New.Live = FALSE)

policy mayUpdate (SELECT *
                  FROM poll AS New, poll AS Old
                  WHERE New.Id = Old.Id
                    AND New.Nam = Old.Nam
                    AND New.Code = Old.Code
                    AND New.Live
                    AND known(Old.Code))

policy sendClient (SELECT *
                   FROM poll
                   WHERE known(poll.Code) OR poll.Live)

type question = int
sequence questionIds
table question : { Id : question, Poll : poll, Text : string }
  PRIMARY KEY Id,
  CONSTRAINT Poll FOREIGN KEY Poll REFERENCES poll(Id)

policy mayInsert (SELECT *
                  FROM question AS New, poll
                  WHERE New.Poll = poll.Id
                    AND known(poll.Code)
                    AND NOT poll.Live
                    AND known(New.Text))

policy sendClient (SELECT *
                   FROM question, poll
                   WHERE question.Poll = poll.Id
                     AND (known(poll.Code) OR poll.Live))

type answers = int
sequence answerIds
table answers : { Id : answers, Poll : poll, Code : int }
  PRIMARY KEY Id,
  CONSTRAINT Poll FOREIGN KEY Poll REFERENCES poll(Id)

policy mayInsert (SELECT *
                  FROM answers AS New, poll
                  WHERE New.Poll = poll.Id
                    AND poll.Live)

policy sendClient (SELECT *
                   FROM answers AS Other, answers AS Self
                   WHERE Other.Poll = Self.Poll
                     AND known(Self.Code))

policy sendClient (SELECT *
                   FROM answers, poll
                   WHERE answers.Poll = poll.Id
                     AND known(answers.Code))

policy sendClient (SELECT *
                   FROM answers, poll, question
                   WHERE answers.Poll = poll.Id
                     AND known(answers.Code)
                     AND question.Poll = poll.Id)

table answer : { Answers : answers, Question : question, Text : string }
  PRIMARY KEY (Answers, Question),
  CONSTRAINT Answers FOREIGN KEY Answers REFERENCES answers(Id),
  CONSTRAINT Question FOREIGN KEY Question REFERENCES question(Id)

policy mayInsert (SELECT *
                  FROM answer AS New, answers, poll, question
                  WHERE New.Answers = answers.Id
                    AND answers.Poll = poll.Id
                    AND question.Poll = poll.Id
                    AND New.Question = question.Id
                    AND known(New.Text)
                    AND poll.Live)

policy sendClient (SELECT *
                   FROM answer, answers, poll
                   WHERE answer.Answers = answers.Id
                     AND answers.Poll = poll.Id
                     AND known(Poll.Code))

policy sendClient (SELECT *
                   FROM answer, answers AS Other, answers AS Self
                   WHERE answer.Answers = Other.Id
                     AND Other.Poll = Self.Poll
                     AND known(Self.Code))

fun wrap title body =
    return <xml>
      <head>
        <title>Poll - {[title]}</title>
      </head>

      <body>
        <h1>{[title]}</h1>

        {body}
      </body>
    </xml>

fun results pid =
    b <- oneRowE1 (SELECT COUNT( * ) > 0
                   FROM poll
                   WHERE poll.Id = {[pid]}
                     AND poll.Live);
    (if b then
         return ()
     else
         error <xml>This poll isn't live yet, so there are no results.</xml>);

    name <- oneRowE1 (SELECT (poll.Nam)
                      FROM poll
                      WHERE poll.Id = {[pid]});
    qs <- queryX1' (SELECT question.Id, question.Text
                    FROM question
                    WHERE question.Poll = {[pid]})
          (fn qu =>
              cs <- queryX1 (SELECT answer.Text
                             FROM answer, answers
                             WHERE answer.Answers = answers.Id
                               AND answers.Poll = {[pid]}
                               AND answer.Question = {[qu.Id]}
                             ORDER BY answer.Text)
                             (fn ch => <xml><li>{[ch.Text]}</li></xml>);
              return <xml>
                <h3>{[qu.Text]}</h3>
                
                {cs}              
              </xml>);
    wrap (name ^ " - Results") qs

fun makePoll r =
    pid <- nextval pollIds;
    code <- rand;
    dml (INSERT INTO poll (Id, Nam, Code, Live)
         VALUES ({[pid]}, {[r.Nam]}, {[code]}, FALSE));
    let
        fun info () =
            ro <- oneOrNoRows1 (SELECT poll.Nam, poll.Live
                                FROM poll
                                WHERE poll.Id = {[pid]}
                                  AND poll.Code = {[code]});
            case ro of
                None => error <xml>Wrong poll admin code!</xml>
              | Some r => return r

        fun resultsForCreator () =
            inf <- info ();
            results pid

        fun goLive () =
            inf <- info ();
            dml (UPDATE poll
                 SET Live = TRUE
                 WHERE Id = {[pid]});
            inf <- info ();
            wrap ("Live poll - " ^ inf.Nam) <xml>
              <a link={resultsForCreator ()}>View latest results</a> (save this URL, as it contains the access code!)
            </xml>

        fun addToPoll () =
            inf <- info ();
            if inf.Live then
                wrap ("Live poll - " ^ inf.Nam) <xml>
                  <a link={resultsForCreator ()}>View latest results</a> (save this URL, as it contains the access code!)
                </xml>
            else
                qs <- queryX1 (SELECT question.Id, question.Text
                               FROM question
                               WHERE question.Poll = {[pid]})
                      (fn qu => <xml><li>{[qu.Text]}</li></xml>);

                wrap ("Non-live poll - " ^ inf.Nam) <xml>
                  <h2>Questions</h2>

                  {qs}

                  <h2>New question</h2>

                  <form>
                    <textbox{#Text}/> <submit value="Add" action={newQuestion}/>
                  </form>

                  <h2>Ready to go live?</h2>
                  
                  <form>
                    <submit value="Yes, go live!" action={goLive}/>
                  </form>
                </xml>

        and newQuestion r =
            inf <- info ();
            if inf.Live then
                error <xml>You can't add a question to a live poll!</xml>
            else
                qid <- nextval questionIds;
                dml (INSERT INTO question (Id, Poll, Text)
                     VALUES ({[qid]}, {[pid]}, {[r.Text]}));
                addToPoll ()
    in
        addToPoll ()
    end

fun vote pid =
    let
        fun info () =
            ro <- oneOrNoRowsE1 (SELECT (poll.Nam)
                                 FROM poll
                                 WHERE poll.Id = {[pid]}
                                   AND poll.Live);
            case ro of
                None => error <xml>That poll doesn't exist or isn't live yet.</xml>
              | Some nam => return nam

        fun castBallot r =
            name <- info ();

            aid <- nextval answerIds;
            code <- rand;
            dml (INSERT INTO answers (Id, Poll, Code)
                 VALUES ({[aid]}, {[pid]}, {[code]}));

            List.app (fn c =>
                         let
                             val qid = readError c.Qid
                         in
                             b <- oneRowE1 (SELECT COUNT( * ) > 0
                                            FROM question
                                            WHERE question.Id = {[qid]}
                                              AND question.Poll = {[pid]});
                             if b then
                                 dml (INSERT INTO answer (Answers, Question, Text)
                                      VALUES ({[aid]}, {[qid]}, {[c.Text]}))
                             else
                                 error <xml>You are mixing questions from different polls!</xml>
                         end) r.Qs;

            let
                fun resultsForVoter () =
                    b <- oneRowE1 (SELECT COUNT( * ) > 0
                                   FROM answers
                                   WHERE answers.Poll = {[pid]}
                                     AND answers.Id = {[aid]}
                                     AND answers.Code = {[code]});

                    if b then
                        results pid
                    else
                        error <xml>Wrong access code for poll results</xml>
            in
                wrap "Thanks for voting!" <xml>
                  You can <a link={resultsForVoter ()}>view the results</a>.  Save this link, as you'll need it to view results later!
                </xml>
            end
    in
        name <- info ();

        qs <- queryX1 (SELECT question.Id, question.Text
                       FROM question
                       WHERE question.Poll = {[pid]})
                      (fn qu => <xml>
                        <entry>
                          <hidden{#Qid} value={show qu.Id}/>
                          <h3>{[qu.Text]}</h3>
                             
                          <textbox{#Text}/>
                        </entry>
                      </xml>);

        wrap name <xml>
          <form>
            <subforms{#Qs}>{qs}</subforms>

            <p><submit value="Vote!" action={castBallot}/></p>
          </form>
        </xml>
    end

fun main () =
    ps <- queryX1 (SELECT poll.Id, poll.Nam
                   FROM poll
                   WHERE poll.Live
                   ORDER BY poll.Nam)
          (fn r => <xml><li><a link={vote r.Id}>{[r.Nam]}</a></li></xml>);
    wrap "Welcome" <xml>
      <form>
        <textbox{#Nam}/>
        <submit value="Create poll" action={makePoll}/>

        <h2>Live Polls</h2>

        {ps}
      </form>
    </xml>
